@@ -3,6 +3,7 @@ import abc
 import enum
 import json
 import time
+import shlex
 import base64
 import string
 import random
@@ -11,6 +12,7 @@ import logging
 import subprocess
 from collections import defaultdict
 import aiohttp
+from aiohttp_socks import ProxyType, ProxyConnector, ChainProxyConnector
 
 
 class AirportLevel(enum.Enum):
@@ -45,7 +47,8 @@ class Node:
         self.test_host = test_host
         self.test_port = test_port
         self.weight = weight
-        self.ping = None
+        self.score = None
+        self.score_unit = 'ms'
         self.protocol = protocol
         self.settings = settings or dict()
         self.stream_settings = stream_settings
@@ -53,10 +56,10 @@ class Node:
 
     @property
     def is_connected(self):
-        return self.ping is not None
+        return self.score is not None
 
     def __str__(self):
-        return f'<Node tag:{self.tag} weight:{self.weight} {self.ping or "--"}ms>'
+        return f'<Node tag:{self.tag} weight:{self.weight} {self.score or "--"}{self.score_unit}>'
 
     def __repr__(self):
         return self.__str__()
@@ -91,6 +94,9 @@ class Airport:
 
 
 class SpeedTest:
+    def __init__(self):
+        self.unit = 'ms'
+
     async def connection_test(self, v):
         task_list = list()
         node_list = list()
@@ -98,10 +104,10 @@ class SpeedTest:
             servers = await airport.pull_node_list()
             for server in servers:
                 node_list.append((airport, server))
-                task = asyncio.create_task(self.try_connect(server))
+                task = asyncio.create_task(self.try_connect(server, v.connect_timeout))
                 task_list.append(task)
         await asyncio.wait(task_list)
-        node_list.sort(key=lambda x: (not x[1].is_connected, -x[1].weight, x[1].ping, x[0].airport_name))
+        node_list.sort(key=lambda x: (not x[1].is_connected, -x[1].weight, x[1].score, x[0].airport_name))
         for airport, node in node_list:
             if node.is_connected:
                 logging.info(f'{node} alive')
@@ -112,15 +118,15 @@ class SpeedTest:
         for airport, node in node_list:
             v.available_airport[airport].add(node)
 
-    async def try_connect(self, node: Node, retry_times: int = 3):
-        node.ping = None
+    async def try_connect(self, node: Node, connect_timeout, retry_times: int = 3):
+        node.score = None
         ping = 9999999
         for i in range(retry_times):
             sleep_time = 3 + i * 2
             try:
                 begin_time = time.time()
                 open_connection = asyncio.open_connection(node.test_host, node.test_port)
-                reader, writer = await asyncio.wait_for(open_connection, timeout=v.connect_timeout)
+                reader, writer = await asyncio.wait_for(open_connection, timeout=connect_timeout)
                 try:
                     await asyncio.sleep(sleep_time)
                     assert not writer.is_closing()
@@ -138,10 +144,12 @@ class SpeedTest:
                 pass
         if ping >= 9999999:
             ping = None
-        node.ping = ping
+        node.score = ping
 
 
 class AvailableTest:
+    TEMP_CONFIG_PREFIX = '/tmp/v2ray_available_test'
+
     CONFIG_TEMPLATE = {
         "inbounds": [
             {
@@ -157,13 +165,14 @@ class AvailableTest:
     }
 
     def __init__(self, http_proxy_port=range(4000, 4008), times=3, response_times=3, sleep_seconds=5,
-                 v2ray_path='/usr/bin/v2ray/v2ray', url_list=None):
+                 v2ray_path='/usr/bin/v2ray/v2ray', url_list=None, unit='ms'):
         self.http_proxy_port = http_proxy_port
         self.times = times
         self.response_times = response_times
         self.sleep_seconds =sleep_seconds
         self.url_list = url_list or ['https://google.com/', 'https://youtube.com/', 'https://github.com/', ]
         self.v2ray_path = v2ray_path
+        self.unit = unit
 
     async def connection_test(self, v):
         task_list = list()
@@ -177,7 +186,7 @@ class AvailableTest:
                 task = asyncio.create_task(self.try_connect(node, port, port_lock[port]))
                 task_list.append(task)
         await asyncio.wait(task_list)
-        node_list.sort(key=lambda x: (not x[1].is_connected, -x[1].weight, x[1].ping, x[0].airport_name))
+        node_list.sort(key=lambda x: (not x[1].is_connected, -x[1].weight, x[1].score, x[0].airport_name))
         for airport, node in node_list:
             if node.is_connected:
                 logging.info(f'{node} stable')
@@ -188,59 +197,137 @@ class AvailableTest:
         for airport, node in node_list:
             v.available_airport[airport].add(node)
 
-    async def try_connect(self, node: Node, port: int, port_lock: asyncio.Lock):
-        node.ping = None
-        ping = 9999999
+    def _init_score(self):
+        if self.unit == 'ms':
+            return 9999999
+        if self.unit == 'Bps':
+            return 0
+        return 9999999
+
+    def _score(self, ms: int, length: int, best: int):
+        if self.unit == 'ms':
+            return min(best, ms)
+        if self.unit == 'Bps':
+            ms = float(ms or 1000)
+            speed = int(length / ms * 1000)
+            return max(speed, best)
+        return best
+
+    async def _popen_connect(self, node: Node, port: int):
+        logger = logging.getLogger()
+        score = self._init_score()
         response_times = 0
+        url = random.choice(self.url_list)
         config = json.loads(json.dumps(self.CONFIG_TEMPLATE))
         config['inbounds'][0]['port'] = port
         config['outbounds'].append(node.to_outbound())
-        config_path = f'/tmp/v2ray_available_test_{port}.json'
+        config_path = f'{self.TEMP_CONFIG_PREFIX}_{port}.json'
+        args = shlex.split(f'"{self.v2ray_path}" "-config" "{config_path}"')
+        p = subprocess.Popen(args)
+        pid = p.pid
+        try:
+            for i in range(self.times):
+                await asyncio.sleep(self.sleep_seconds)
+                try:
+                    begin_time = time.time()
+                    async with aiohttp.ClientSession() as session:
+                        async with session.get(url=url, timeout=5, proxy=f"http://127.0.0.1:{port}") as resp:
+                            text = await resp.text()
+                    length = len(text or '')
+                    response_times += 1
+                    finish_time = time.time()
+                    current_ping = int((finish_time - begin_time) * 1000)
+                    logger.info(f'{node} times: {i + 1} score:{current_ping}ms response: {length} bytes')
+                    score = self._score(current_ping, length, score)
+                except Exception as e:
+                    logger.error(f'{node} available test failed: {e}')
+                    if self.times - i - 1 + response_times < self.response_times:
+                        break
+        finally:
+            p.kill()
+            os.popen(f'kill {pid}')
+            os.remove(config_path)
+            p.communicate()
+        return score, response_times
+
+    async def _http_connect(self, node: Node):
+        logger = logging.getLogger()
+        score = self._init_score()
+        response_times = 0
         url = random.choice(self.url_list)
-        async with port_lock:
-            with open(config_path, "w") as f:
-                f.write(json.dumps(config, indent=2))
-            assert os.path.exists(config_path)
-            import shlex
-            args = shlex.split(f'"{self.v2ray_path}" "-config" "{config_path}"')
-            p = subprocess.Popen(args)
-            pid = None
+        host = node.settings['servers'][0]['address']
+        port = node.settings['servers'][0]['port']
+        for i in range(self.times):
+            await asyncio.sleep(self.sleep_seconds)
             try:
-                pid = p.pid
-                for i in range(self.times):
-                    await asyncio.sleep(self.sleep_seconds)
-                    try:
-                        begin_time = time.time()
-                        async with aiohttp.ClientSession() as session:
-                            async with session.get(url=url, timeout=5, proxy=f"http://127.0.0.1:{port}") as resp:
-                                await resp.text()
-                        response_times += 1
-                        finish_time = time.time()
-                        ping = min(int((finish_time - begin_time) * 1000), ping)
-                    except Exception as e:
-                        logging.getLogger().error(f'{node} available test failed: {e}')
-                        if self.times - i - 1 + response_times < self.response_times:
-                            break
-            finally:
-                p.kill()
-                os.remove(config_path)
-                if pid is not None:
-                    os.system(f'kill -9 {pid}')
+                begin_time = time.time()
+                async with aiohttp.ClientSession() as session:
+                    async with session.get(url=url, timeout=5, proxy=f"http://{host}:{port}") as resp:
+                        text = await resp.text()
+                length = len(text or '')
+                response_times += 1
+                finish_time = time.time()
+                current_ping = int((finish_time - begin_time) * 1000)
+                logger.info(f'{node} times: {i + 1} score:{current_ping}ms response: {length} bytes')
+                score = self._score(current_ping, length, score)
+            except Exception as e:
+                logger.error(f'{node} available test failed: {e}')
+                if self.times - i - 1 + response_times < self.response_times:
+                    break
+        return score, response_times
+
+    async def _socks_connect(self, node: Node):
+        logger = logging.getLogger()
+        score = self._init_score()
+        response_times = 0
+        url = random.choice(self.url_list)
+        host = node.settings['servers'][0]['address']
+        port = node.settings['servers'][0]['port']
+        for i in range(self.times):
+            await asyncio.sleep(self.sleep_seconds)
+            try:
+                begin_time = time.time()
+                connector = ProxyConnector.from_url(f'socks5://{host}:{port}')
+                async with aiohttp.ClientSession(connector=connector) as session:
+                    async with session.get(url=url, timeout=5) as resp:
+                        text = await resp.text()
+                length = len(text or '')
+                response_times += 1
+                finish_time = time.time()
+                current_ping = int((finish_time - begin_time) * 1000)
+                logger.info(f'{node} times: {i + 1} score:{current_ping}{self.unit} response: {length} bytes')
+                score = self._score(current_ping, length, score)
+            except Exception as e:
+                logger.error(f'{node} available test failed: {e}')
+                if self.times - i - 1 + response_times < self.response_times:
+                    break
+        return score, response_times
+
+    async def try_connect(self, node: Node, port: int, port_lock: asyncio.Lock):
+        node.score = None
+        if node.protocol == 'http':
+            ping, response_times = await self._http_connect(node)
+        elif node.protocol == 'socks':
+            ping, response_times = await self._socks_connect(node)
+        else:
+            async with port_lock:
+                ping, response_times = await self._popen_connect(node, port)
         if ping >= 9999999:
             ping = None
         if response_times < self.response_times:
             ping = None
-        node.ping = ping
+        node.score = ping
+        node.score_unit = self.unit
 
 
 class V2Ray:
     def __init__(self,
                  config_path: str = '/etc/v2ray/config.json',
-                 config_template: str = None,
+                 config_template: dict = None,
                  restart_command: str = 'service v2ray restart',
                  connect_timeout: int = 4,
                  period_seconds: int = 8 * 3600,
-                 ping_latency_ms: int = None,
+                 score_limit: int = None,
                  test_object = None,
                  config_object = None,
                  ):
@@ -253,7 +340,7 @@ class V2Ray:
         self.period_seconds = period_seconds
         self.template = config_template
         self.restart_command = restart_command
-        self.ping_latency_ms = ping_latency_ms
+        self.score_limit = score_limit
         self.test_object = test_object or SpeedTest()
         self.config_object = config_object or OutboundConfigBuilder()
 
@@ -265,11 +352,18 @@ class V2Ray:
             available_airport = {airport: node_set for airport, node_set in self.available_airport.items()
                                  if airport.level != AirportLevel.SOS}
         for airport, node_set in available_airport.items():
-            node_list = sorted(node_set, key=lambda x: x.ping)
+            if self.test_object.unit == 'Bps':
+                node_list = sorted(node_set, key=lambda x: x.score, reverse=True)
+            else:
+                node_list = sorted(node_set, key=lambda x: x.score)
             node_list = node_list[:airport.max_devices]
             for node in node_list:
-                if self.ping_latency_ms and node.ping > self.ping_latency_ms:
-                    continue
+                if self.test_object.unit == 'Bps':
+                    if self.score_limit and node.score < self.score_limit:
+                        continue
+                else:
+                    if self.score_limit and node.score > self.score_limit:
+                        continue
                 nodes.append(node)
         return nodes
 
